

package com.aelitis.azureus.core.peermanager.unchoker;

import java.util.*;

import org.gudy.azureus2.core3.peer.PEPeer;
import org.gudy.azureus2.core3.peer.impl.PEPeerTransport;
import org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol;



/**
 * 
 * An unchoker that returns proportional shares to other peers.
 * Based on the BitTyrant's JarUnchoker
 * 
 * @author DaveLevin
 * 
 */
public class PropShareUnchoker extends GlobalTorrentAwareDownloadingUnchoker {

	class TopPropPrevRoundCompare implements Comparator<PEPeerTransportProtocol>
	{
		long roundDuration;
		TopPropPrevRoundCompare(long _roundDuration)
		{
			this.roundDuration = _roundDuration;
		}
		public int compare(PEPeerTransportProtocol a, PEPeerTransportProtocol b) 
		{
			double a_score = (double) (a.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2)) /
				(a.getTyrantStats().getDataBytesSentInTheLast(2*roundDuration + 2) - a.getTyrantStats().getDataBytesSentInTheLast(roundDuration + 2) + 1);
			double b_score = (double) (b.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2)) /
				(b.getTyrantStats().getDataBytesSentInTheLast(2*roundDuration + 2) - b.getTyrantStats().getDataBytesSentInTheLast(roundDuration + 2) + 1);

			if(a_score == b_score)
				return 0;
			if(a_score > b_score)
				return 1;
			return -1;
		}
	}

	class TopPropHistoricCompare implements Comparator<PEPeerTransportProtocol>
	{
		public int compare(PEPeerTransportProtocol a, PEPeerTransportProtocol b) 
		{
			double a_score = (double) (a.getStats().getTotalDataBytesReceived()) /
				(a.getStats().getTotalDataBytesSent() + 1);
			double b_score = (double) (b.getStats().getTotalDataBytesReceived()) /
				(b.getStats().getTotalDataBytesSent() + 1);

			if(a_score == b_score)
				return 0;
			if(a_score > b_score)
				return 1;
			return -1;
		}
	}

	long lastTotalAllocation = 0;

	public PropShareUnchoker() { }



	/**
	 * What should we do if we find ourselves in a situation where we've
	 * allocated enough bandwidth to everyone to get reciprication, but we still
	 * have more upload capacity? Set to true to help the swarm by evenly
	 * distributing any leftover bandwidth amungst all unchoked peers.
	 */
	public static boolean EVENLY_ALLOCATE_EXCESS_BANDIWDTH = true;


	// what do we say is the least acceptable amount to give to a peer?
	// this is number of bytes per a 10 second allocation
	public static long ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER = 30000;
	//public static long ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER = 120000;

	public static double RATE_IN_TO_RATE_OUT_CONVERSION = 2.0;

	private static ChokesAndUnchokes cache = null;

	private static long lastCalled = 0;

	private static long roundDuration = 10;

	private TopPropHistoricCompare compare_historic = new TopPropHistoricCompare();
	private TopPropPrevRoundCompare compare_prev_round = new TopPropPrevRoundCompare(roundDuration);

	// Fraction of bandwidth to allocate to prop share; the rest will go
	// to learning about new peers, etc.
	private static double PROP_SHARE_BUDGET_FRAC = .8;
	private static double MAX_FUDGE_LEARN_FRAC = .1;	// when reverse-engineering best point, allow this much jitter
	private static long NUM_ROUNDS_FOR_BW_CAP_EST = 5;
	private static double MAX_BW_EST_FACTOR = 2.0; // how many times more than the max pulled we're willing to give

	private long best_allocation(PEPeerTransportProtocol pep, long uploadBandwidthRemaining, long fallbackAllocation)
	{
		// We have uploadBandwidthRemaining left; how much of that (if any) should we give to nbr?
		// Option 1: a constant amount to each
		// Option 2: whatever gave us the best response historically
		// Option 3: "reverse engineer" their response and optimize that
		if(pep.getTyrantStats().getHistory().size() == 0)
		{
			// we have no information; try the fallback allocation
			return fallbackAllocation;
		}

		long expected_return = pep.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2);
		long rv = pep.getTyrantStats().getDataBytesSentInTheLast(2*roundDuration + 2) - pep.getTyrantStats().getDataBytesSentInTheLast(roundDuration + 2);
		for(HistoricalDataPoint h : pep.getTyrantStats().getHistory())
		{
			long taken = h.getDataReceivedFromPeer();
			long given = h.getDataSentToPeer();

			if(given == 0) continue; // might have been optimistically unchoked
			if((taken > expected_return) && (given < rv)) // clearly better
			{
				expected_return = taken;
				rv = given;
			}
			else if(taken < expected_return && given > rv) // clearly worse
				continue;
			else if(rv == 0)
			{
				if(taken > expected_return && given <= uploadBandwidthRemaining)
				{
					// we have bw to blow; might as well spend it for better profit
					expected_return = taken;
					rv = given;
				}
			}
			else
			{
				// both denominators are non-zero due to prior else's
				double delta_recv = taken / expected_return;
				double delta_send = given / rv;
				if(delta_recv / delta_send > 1) // better return for paying less
				{
					expected_return = taken;
					rv = given;
				}
				else // we seem to have gotten penalized
					continue;
			}
		}

		// Randomly fudge it for learning's sake
		double fudge = Math.random() * MAX_FUDGE_LEARN_FRAC;
		if(Math.random() < 0.5)
			rv *= (1.0+fudge);
		else
			rv *= (1.0-fudge);

		return rv;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.aelitis.azureus.core.peermanager.unchoker.GlobalTorrentAwareDownloadingUnchoker#calculateChokesAndUnchokes(java.util.ArrayList,
	 *      long)
	 */

	@Override
	public ChokesAndUnchokes calculateChokesAndUnchokes(ArrayList all_peers, long uploadBandwidthToAllocate) 
	{
		// we're caching results because will be called multiple times each 10
		// sec if more than 1 torrent is downloading
		if (cache != null && lastCalled + 5000 > System.currentTimeMillis()) 
		{
			System.out.println("hit cache");
			return cache;
		}

		// update histories of everyone
		for (PEPeerTransportProtocol pep : (List<PEPeerTransportProtocol>) all_peers) 
		{
			if (pep.getStats() != null) 
			{
				long sent = pep.getStats().getTotalDataBytesSent();
				long received = pep.getStats().getTotalDataBytesReceived();
				long alloc = pep.getTyrantStats().getLastUploadAllocation();
				long pulled = alloc - pep.getNetworkConnection().getBandwidthLimit();
				pep.getTyrantStats().updateHistory(new HistoricalDataPoint(sent, received, alloc, pulled));
			}
		}


		long total_given_for_free = 0;
		long total_given_in_return = 0;
		
		long totalBandwidthAllotted = uploadBandwidthToAllocate;

		// long realLimit = uploadBandwidthToAllocate; //keep this limit around
		// because we're going to mess with uploadBandwidthToAllocate
		ArrayList chokes = new ArrayList();
		ArrayList unchokes = new ArrayList();

		System.out.println("allocating " + uploadBandwidthToAllocate + " bytes for the next 10 sec");

		// figure out how much we used last round and how much we got last round
		long unusedBandwidthLastRound = 0;
		long allocatedBandwidthLastRound = 0;
		long unusedBandwidthFromFlippedPeers = 0;
		int numUnchoked = 0;
		int numPeersThatFlipped = 0;
		long bw_recvd_last_round = 0;
		long weighted_contribution = 0;
		for (PEPeerTransportProtocol pet : (List<PEPeerTransportProtocol>) all_peers) 
		{
			long cont_1 = pet.getTyrantStats().getDataBytesReceivedInTheLast(1*roundDuration + 2);
			long cont_2 = pet.getTyrantStats().getDataBytesReceivedInTheLast(2*roundDuration + 2) - cont_1;
			long cont_3 = pet.getTyrantStats().getDataBytesReceivedInTheLast(3*roundDuration + 2) - cont_2 - cont_1;
			long cont_4 = pet.getTyrantStats().getDataBytesReceivedInTheLast(4*roundDuration + 2) - cont_3 - cont_2 - cont_1;

			long wt_cont = (long) (cont_1 + (.5 * cont_2) + (.25 * cont_3) + (.125 * cont_4));

			// DEBUG
			System.out.println("[PROPSHARE] Peer " + pet + "'s last 4 contributions = [ " + cont_4 + " " + cont_3 + " " + cont_2 + " " + cont_1 + " ] wt-cont: " + wt_cont);


			if (UnchokerUtil.isUnchokable(pet, true))
			{
				bw_recvd_last_round += pet.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2);
				weighted_contribution +=  wt_cont;
					//(.25* pet.getTyrantStats().getDataBytesReceivedInTheLast(3*roundDuration) +
					 //.25* pet.getTyrantStats().getDataBytesReceivedInTheLast(2*roundDuration) +
					 //.5* pet.getTyrantStats().getDataBytesReceivedInTheLast(1*roundDuration));

				System.out.println("[PROPSHARE] Counting " + pet + "'s recent contribution of BW " +
					       pet.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2) + 
					       " WBW " + wt_cont + " - he is unchokable");
			}
			else
			{
				System.out.println("[PROPSHARE] Not counting " + pet + "'s contribution of BW " + 
					       pet.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2) + 
					       " WBW " + wt_cont + " - he is not unchokable");
			}

			if (!pet.isChokedByMe()) 
			{
				unusedBandwidthLastRound += pet.getNetworkConnection().getBandwidthLimit();
				allocatedBandwidthLastRound += pet.getTyrantStats().getLastUploadAllocation();
				numUnchoked++;
				if (!pet.isInterested()) 
				{
					numPeersThatFlipped++;
					unusedBandwidthFromFlippedPeers += pet.getNetworkConnection().getBandwidthLimit();
				}
				System.out.println("[PROPSHARE-DEBUG] " + System.currentTimeMillis() + " " + pet + 
						" was given " + pet.getTyrantStats().getLastUploadAllocation() +
						" but has " + pet.getNetworkConnection().getBandwidthLimit() + " left over" +
						" used frac = " + (1.0 - ((double)pet.getNetworkConnection().getBandwidthLimit() / pet.getTyrantStats().getLastUploadAllocation()))); 

				// Update the bandwidth cap based on historical data
				long max_pulled = pet.getTyrantStats().getMaxDataBytesPulledInTheLast((NUM_ROUNDS_FOR_BW_CAP_EST*roundDuration)+2);
				long bw_cap = Math.max(ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER, (long) max_pulled);

				bw_cap *= MAX_BW_EST_FACTOR;

				pet.getTyrantStats().setBandwidthCap(bw_cap);
				System.out.println("[PROPSHARE-DEBUG] " + System.currentTimeMillis() + 
						" giving a bandwidth cap of " + bw_cap + " to " + pet);
			}
		}

		double percent = (double) unusedBandwidthLastRound / allocatedBandwidthLastRound;
		System.out.println("Allocation percent unused last round: " + percent);
		System.out.println(numPeersThatFlipped + " peers out of " + numUnchoked + " peers went from interested to uninterested in the last round");
		System.out.println(unusedBandwidthFromFlippedPeers + " bytes were left allocated to the peers that flipped");
		System.out.println("[PROPSHARE] " + System.currentTimeMillis() + " total bw received last round = " + bw_recvd_last_round);
		System.out.println("[PROPSHARE] " + System.currentTimeMillis() + " total weighted contribution = " + weighted_contribution);

		int fallbackAllocation = 120000; // that's 3k/sec

		// similar to market rate... what to give peers if we don't know
		// put at the front of the list who we think is the best to download
		// from
		Collections.sort((List<PEPeerTransport>) all_peers);


		/****************************************************************************************************/
		/****************************************************************************************************/
		// MAIN PROPSHARE LOOP
		// Run twice, for snubbed and unsnubbed peers
		/****************************************************************************************************/
		/****************************************************************************************************/
		
		double bw_to_allocate_to_prop = totalBandwidthAllotted * PROP_SHARE_BUDGET_FRAC;

		for (boolean allow_snubbed : new Boolean[] { false, true }) 
		{
			Iterator pit = all_peers.iterator();
			while (pit.hasNext() && uploadBandwidthToAllocate > 0) 
			{
				PEPeerTransportProtocol pep = (PEPeerTransportProtocol) pit.next();

				long cont_1 = pep.getTyrantStats().getDataBytesReceivedInTheLast(1*roundDuration + 2);
				long cont_2 = pep.getTyrantStats().getDataBytesReceivedInTheLast(2*roundDuration + 2) - cont_1;
				long cont_3 = pep.getTyrantStats().getDataBytesReceivedInTheLast(3*roundDuration + 2) - cont_2 - cont_1;
				long cont_4 = pep.getTyrantStats().getDataBytesReceivedInTheLast(4*roundDuration + 2) - cont_3 - cont_2 - cont_1;

				long wt_contribution = (long) (cont_1 + (.5 * cont_2) + (.25 * cont_3) + (.125 * cont_4));
				
				long bw_contribution = pep.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2);
				if(bw_recvd_last_round > 0)
				{
					System.out.println("[PROPSHARE] If I were sending based on BW, " + pep + " would have gotten " + 
							(long)((bw_contribution * bw_to_allocate_to_prop) / bw_recvd_last_round));
				}
				
				if(UnchokerUtil.isUnchokable(pep, allow_snubbed) && (wt_contribution > 0))
				{
					// keep them
					long toAllocate = fallbackAllocation;

					if(weighted_contribution > 0)
					{
						toAllocate = (long)((wt_contribution * bw_to_allocate_to_prop) / weighted_contribution);
					}

					if(toAllocate > RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution)
					{
						System.out.println("[RATE-CAP] Would have given " + toAllocate + 
								" but will instead cap at " + (RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution) + " for peer " + pep);
						toAllocate = (long) (RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution);
					}

					if(toAllocate < ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER)
						toAllocate = ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER;
					if(toAllocate > uploadBandwidthToAllocate)
						toAllocate = uploadBandwidthToAllocate;
					if(toAllocate > ((PEPeerTransportProtocol) pep).getTyrantStats().getBandwidthCap())
						toAllocate = ((PEPeerTransportProtocol) pep).getTyrantStats().getBandwidthCap();
					
					((PEPeerTransportProtocol) pep).getTyrantStats().setLastUploadAllocation((int) toAllocate);
					uploadBandwidthToAllocate -= toAllocate;
					System.out.println("[PASS 1] - kept " + pep + " because they sent to us");
					pep.getTyrantStats().setLastTyrantAction("Decreased upload allocation to honor prop share.");

					// assign and print
					pep.getNetworkConnection().setBandwidth((int) toAllocate);
					System.out.println("[PASS 1] - " + pep + " got " + toAllocate + " bytes for the next 10 sec");
					total_given_in_return += toAllocate;

					unchokes.add(pep);
					pit.remove();
				}
			}

			// drain iterator
			while (pit.hasNext())
				pit.next();

			/****************************************************************************************************/
			// Research; do some minimal optimistic unchoking, basically
			/****************************************************************************************************/


			// We know that for each of these peers wt_contribution == 0

			if(uploadBandwidthToAllocate > 0)
			{
				//Collections.sort((List<PEPeerTransportProtocol>) all_peers, compare_prev_round);
				Collections.sort((List<PEPeerTransport>) all_peers);
				pit = all_peers.iterator();
				while (pit.hasNext() && uploadBandwidthToAllocate > 0)
				{
					PEPeerTransportProtocol pep = (PEPeerTransportProtocol) pit.next();
					// allow_snubbed was false
					if (UnchokerUtil.isUnchokable(pep, allow_snubbed))
					{
						long toAllocate = fallbackAllocation;

						if(toAllocate < ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER)
							toAllocate = ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER;
						if(toAllocate > uploadBandwidthToAllocate)
							toAllocate = uploadBandwidthToAllocate;
						if(toAllocate > ((PEPeerTransportProtocol) pep).getTyrantStats().getBandwidthCap())
							toAllocate = ((PEPeerTransportProtocol) pep).getTyrantStats().getBandwidthCap();

						((PEPeerTransportProtocol) pep).getTyrantStats().setLastUploadAllocation(toAllocate);
						uploadBandwidthToAllocate -= toAllocate;
						System.out.println("[PASS 2] - kept " + pep + " as research; they've not been good to us lately, but we'd like them to be");
						pep.getTyrantStats().setLastTyrantAction("Decreased upload allocation because he gave well historically.");

						// assign and print
						pep.getNetworkConnection().setBandwidth((int) toAllocate);
						System.out.println("[PASS 2] - " + pep + " got " + toAllocate + " bytes for the next 10 sec");
						total_given_for_free += toAllocate;

						unchokes.add(pep);
						pit.remove();
					}
				}
			}


			/****************************************************************************************************/
			/****************************************************************************************************/
			// We were short-changing our prop-share before to try to do some research; give back
			/****************************************************************************************************/
			/****************************************************************************************************/



			pit = unchokes.iterator(); // NOTE we're iterating through unchokes, whereas we normally iterate through all_peers
			bw_to_allocate_to_prop = uploadBandwidthToAllocate; // give out via prop share all that we have left
			while(pit.hasNext() && uploadBandwidthToAllocate > 0)
			{
				PEPeerTransportProtocol pep = (PEPeerTransportProtocol) pit.next();

				long contribution = pep.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2);

				long cont_1 = pep.getTyrantStats().getDataBytesReceivedInTheLast(1*roundDuration + 2);
				long cont_2 = pep.getTyrantStats().getDataBytesReceivedInTheLast(2*roundDuration + 2) - cont_1;
				long cont_3 = pep.getTyrantStats().getDataBytesReceivedInTheLast(3*roundDuration + 2) - cont_2 - cont_1;
				long cont_4 = pep.getTyrantStats().getDataBytesReceivedInTheLast(4*roundDuration + 2) - cont_3 - cont_2 - cont_1;

				long wt_contribution = (long) (cont_1 + (.5 * cont_2) + (.25 * cont_3) + (.125 * cont_4));

				long bw_contribution = pep.getTyrantStats().getDataBytesReceivedInTheLast(roundDuration + 2);
				if(bw_recvd_last_round > 0)
				{
					System.out.println("[PROPSHARE] If I were sending based on BW, " + pep + " would have gotten " + 
							(long)((bw_contribution * bw_to_allocate_to_prop) / bw_recvd_last_round) + " more");
				}
				
				if(UnchokerUtil.isUnchokable(pep, allow_snubbed) && (wt_contribution > 0))
				{
					// keep them
					long moreToAllocate = fallbackAllocation;

					if(weighted_contribution > 0)
					{
						moreToAllocate = (long)((wt_contribution * bw_to_allocate_to_prop) / weighted_contribution);
					}

					if(wt_contribution == 0)
					{
						System.out.println("[FAILED-ASSERTION] " + pep + " has wt_contribution = 0 but should not in PASS 3");
						continue;
					}
				
					System.out.println("[SANITY-CHECK] Last upload allocation to " + pep + " = " + pep.getTyrantStats().getLastUploadAllocation());
					long toAllocate = moreToAllocate + pep.getTyrantStats().getLastUploadAllocation(); // The difference from pass 1

					if(toAllocate > RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution)
					{
						System.out.println("[RATE-CAP] Would have given " + toAllocate + 
								" but will instead cap at " + (RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution) + " for peer " + pep);
						toAllocate = (long) (RATE_IN_TO_RATE_OUT_CONVERSION * wt_contribution);
					}

					if(toAllocate < ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER)
						toAllocate = ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER;
					if(toAllocate > uploadBandwidthToAllocate)
						toAllocate = uploadBandwidthToAllocate;
					if(toAllocate > pep.getTyrantStats().getBandwidthCap())
						toAllocate = pep.getTyrantStats().getBandwidthCap();

					pep.getTyrantStats().setLastUploadAllocation((int) toAllocate);
					uploadBandwidthToAllocate -= moreToAllocate;
					System.out.println("[PASS 3] - kept " + pep + " got " + moreToAllocate + " more for a total of " + toAllocate);
					pep.getTyrantStats().setLastTyrantAction("[PASS 3] - Gave some more to a prop-share peer; we didn't spend our whole research budget.");

					// assign and print
					pep.getNetworkConnection().setBandwidth((int) toAllocate);
					System.out.println("[PASS 3] - " + pep + " got " + toAllocate + " bytes for the next 10 sec");
					total_given_in_return += moreToAllocate;

					// Don't do this now; we are iterating through unchokes
					// unchokes.add(pep);
					// pit.remove();
				}
			}

			// drain iterator
			while (pit.hasNext())
				pit.next();

		}

		/****************************************************************************************************/
		/****************************************************************************************************/
		// Make chokes and set connection weights
		/****************************************************************************************************/
		/****************************************************************************************************/
		
		// make chokes
		for (Object o : all_peers) {
			if (!unchokes.contains(o)) {
				chokes.add(o);
				((PEPeer) o).getTyrantStats().setLastTyrantAction("choked");
			}
		}

		// give everyone else bootstrap bandwidth so we can send
		// interested/KA/have messages
		for (PEPeerTransportProtocol pep : (List<PEPeerTransportProtocol>) chokes) {
			if (pep.getNetworkConnection() != null)
				pep.getNetworkConnection().setBandwidth((int) ABSOLUTE_MINIMUM_ALLOCATION_PER_PEER);
		}

		// figure out a weight for each unchoked connection
		// XXX Use it or lose it!
		long totalUploadAllocated = 0;
		for (PEPeerTransportProtocol pep : (List<PEPeerTransportProtocol>) unchokes) {
			totalUploadAllocated += pep.getTyrantStats().getLastUploadAllocation();
		}
		// TODO Look into this - what is the weight all about
		for (PEPeerTransportProtocol pep : (List<PEPeerTransportProtocol>) unchokes) {
			double connectionWeight = ((double)pep.getTyrantStats().getLastUploadAllocation());
			pep.getNetworkConnection().setWeight(connectionWeight);
			
			System.out.println("weight of " + connectionWeight + " assigned to " + pep.getNetworkConnection());
		}

		System.out.println("[PROPSHARE-GIVEN-STATS] " + System.currentTimeMillis() + 
				" givenForFree: " + total_given_for_free +
				" givenInReturn: " + total_given_in_return);

		lastCalled = System.currentTimeMillis();
		cache = new ChokesAndUnchokes(chokes, unchokes);

		return cache;
	}
}
